/*
 * LispParser.java
 *
 * Created on February 26, 2002, 6:08 PM
 */

package org.semanticweb.daml;

import java.util.*;
import java.io.*;

/**
 * This is a simple LISP parser for parsing output of some
 * DL classifiers. LISP lists are represented as Java Lists.
 * @author  Michael Sintek
 * @version 
 */
public class LispParser {

  static final int EOF = -1;

  public static Object parseExpression(Reader reader) throws Exception {
    // returns:
    // - null if no expression was read
    // - Java List for a list (empty Java List for nil, not null!)
    // - String for symbols & strings
    // - (not implemented: Integer/Double for numbers)
    // expression ::= list | symbol
    skipWhitespace(reader);
    reader.mark(1);
    int c = reader.read();
    reader.reset();
    if (c == EOF)      return null;
    else if (c == '(') return parseList(reader);
    else if (c == ')') return null;
    else if (c == '"') return parseSymbol(reader, '"');
    else if (c == '|') return parseSymbol(reader, '|');
    else               return parseSymbol(reader);
  }
  
  public static Object parseExpressionList(Reader reader) throws Exception {
    // expr* EOF
    ArrayList list = new ArrayList();
    Object expr = parseExpression(reader);
    while (expr != null) {
      list.add(expr);
      expr = parseExpression(reader);
    }
    accept(reader, EOF);
    return list;
  }
  
  static void skipWhitespace(Reader reader) throws Exception {
    reader.mark(1);
    int c = reader.read();
    if (c == ';') { // comment until EOL
      while (c != '\n' && c != '\r' && c != EOF)
        c = reader.read();
      skipWhitespace(reader);
    } else if (Character.isWhitespace((char)c)) {
      skipWhitespace(reader);
    } else
      reader.reset();
  }

  static String charString(int c) {
    if (c == EOF) 
      return "EOF";
    else if (!Character.isISOControl((char)c)) 
      return String.valueOf((char)c);
    else 
      return "#" + c;
  }
  
  static void accept(Reader reader, int exp) throws Exception {
    int c = reader.read();
    if (c != exp)
      throw new Exception("expected " + charString(exp) 
                          + ", got " + charString(c));
  }
  
  public static Object parseList(Reader reader) throws Exception {
    // '(' expr* ')'
    ArrayList list = new ArrayList();
    accept(reader, '(');
    Object expr = parseExpression(reader);
    while (expr != null) {
      list.add(expr);
      expr = parseExpression(reader);
    }
    accept(reader, ')');
    return list;
  }
  
  public static Object parseSymbol(Reader reader, int delimiter) throws Exception {
    // symbol delimited by some character (" or |)
    accept(reader, delimiter);
    StringBuffer buffer = new StringBuffer();
    reader.mark(1);
    int c = reader.read();
    while (c != delimiter && c != EOF) {
      buffer.append((char)c);
      reader.mark(1);
      c = reader.read();
    }
    reader.reset();
    accept(reader, delimiter);
    return buffer.toString();
  }
  
  public static Object parseSymbol(Reader reader) throws Exception {
    // normal LISP symbol
    StringBuffer buffer = new StringBuffer();
    reader.mark(1);
    int c = reader.read();
    while (   !Character.isWhitespace((char)c)
           && c != '(' && c != ')' && c != ';'
           && c != '"' && c != '|' && c != EOF) {
      buffer.append((char)c);
      reader.mark(1);
      c = reader.read();
    }
    reader.reset();
    String symbol = buffer.toString();
    if (symbol.equalsIgnoreCase("nil"))
      return new ArrayList(0);
    else
      return buffer.toString(); // check if it is a number ... !!!
  }
  
  // testing
  
  public static void main(String[] args) throws Exception {
    System.out.println(parseExpressionList(
      new BufferedReader(new FileReader(args[0]))));
  }
  
}

