/*
 * Commandline.java
 *
 * Created on February 8, 2002, 9:57 AM
 */

package org.semanticweb.triple.application;

import java.io.*;

import org.w3c.rdf.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.implementation.model.*;
import org.w3c.rdf.syntax.*;
import org.xml.sax.InputSource;
import com.hp.hpl.jena.rdf.arp.StanfordImpl;

import org.semanticweb.triple.parser.*;
import org.semanticweb.triple.language.Clause;
import org.semanticweb.glp.*;
import org.semanticweb.lp.*;


/**
 * Process TRIPLE programs from command line.
 * @author  Michael Sintek
 * @version
 */
public class Commandline {
  
  final static String OUT = "#out.P";
  final static String XSBHEADER = "/xsb/header.P";
  final static String XSBTRAILER = "/xsb/trailer.P";
  
  public static void compileTripleProgram(String inputFileName, 
      String outputFileName, String dir, String timestamp, boolean servlet) 
      throws Exception {
    compileTripleProgram(inputFileName, outputFileName, dir, timestamp, servlet,
      null, null, null, null, null, null);
  }

  public static void compileTripleProgram(String inputFileName, String outputFileName,
      String dir, String timestamp, // for further temp files
      boolean servlet,
      String rdfFileName, String rdfNamespaceString, String rdfNameString,
      String rdfDumpFileName, String rdfDumpNamespaceString, 
      String rdfDumpNameString) throws Exception {
    LTermParser parser = null;
    if (inputFileName == null) {
      parser = new LTermParser(System.in);
      inputFileName = "standard input";
    } else {
      try {
        parser = new LTermParser(new java.io.FileInputStream(inputFileName));
      } catch (java.io.FileNotFoundException e) {
        System.err.println("Error: file '" + inputFileName + "' not found.");
        System.exit(1);
      }
    }
    PrintWriter writer = new PrintWriter(new FileWriter(outputFileName));
    writer.println("% " + outputFileName + ", generated from " + inputFileName + "\n");
    writer.println("triple_dir_timestamp('" + dir + "','" + timestamp + "').");
    if (servlet)
      writer.println("triple_allow_external('daml_oil').\n"); // pathname? ###
    else
      writer.println("triple_allow_external(_).\n");
    copyFile(writer, XSBHEADER);
    ExternalPrologCallback callback = new ExternalPrologCallback(writer);
    parser.parse(new Context(callback));
    // handle -rdfdata
    if (rdfFileName != null) {
       callback.setRDFNamespace(new ResourceImpl(rdfNamespaceString,rdfNameString));
       compileRDFFile(rdfFileName,callback,rdfNamespaceString);
    }
    // handle -rdfdump
    if (rdfDumpFileName != null) {
      writer.println(":- writeln('***').");
      writer.println(":- writeln('dumping model as RDF file ...').");
      writer.println(":- dump_model('" + rdfDumpFileName + "', resource('" +
        rdfDumpNamespaceString + "','" + rdfDumpNameString + "')).");
      writer.println(":- writeln('done.').\n\n");
    }
    copyFile(writer, XSBTRAILER);
    writer.close();
    System.err.println("'" + outputFileName + "' written.");
  }

  static void copyFile(PrintWriter writer, String fileName) throws Exception {
    InputStream is = new Commandline().getClass().getResourceAsStream(fileName);
    if (is != null) {
      BufferedReader reader = new BufferedReader(new InputStreamReader(is));
      String line = reader.readLine();
      while (line != null) {
        writer.println(line);
        line = reader.readLine();
      }
      writer.println();
      reader.close();
    } else {
      System.err.println("resource '" + fileName + "' not found");
      System.err.println("[make sure java/resources is on your classpath]");
    }
  }
  
  static void compileRDFFile (String filename, ExternalPrologCallback callback, String id) throws Exception{
    compileRDFStream (new FileInputStream (filename),filename, callback,id);
  }
 
  static void compileRDFStream (InputStream rdfdstream, String ResourceName, 
      ExternalPrologCallback callback, String id) throws Exception {
    //RDFParser parser = new RDFFactoryImpl().createParser(); 
    RDFParser parser = new com.hp.hpl.jena.rdf.arp.StanfordImpl();
    InputSource input = new InputSource(rdfdstream);
    input.setSystemId(id);
    input.setPublicId(id);
    parser.parse(input, callback);
    //System.err.println("Loaded " + new Integer(count) + " Triples.");
  }


  public static void main(String args[]) {
    int i = 0;
    String inputFileName = null;
    // for -rdfdata
    String rdfFileName = null;
    String rdfNamespaceString = null;
    String rdfNameString = null;
    // for -rdfdump
    String rdfDumpFileName = null;
    String rdfDumpNamespaceString = null;
    String rdfDumpNameString = null;

    if (args.length == 0) {
      System.err.println("arguments expected: [-rdfdata <rdf-file> <namespace> <localname>]");
      System.err.println("                    [-rdfdump <rdf-file> <namespace> <localname>]");
      System.err.println("                    (<filename>|-)");
      System.err.println(" -rdfdata: load RDF file into specified model");
      System.err.println(" -rdfdump: dump specified model as RDF file");
      System.err.println(" -       : read from standard input");
      System.exit(1);
    } else {
      if (args.length >= i+4 && args[i].equalsIgnoreCase("-rdfdata")) {
        rdfFileName = args[i+1];
        rdfNamespaceString = args[i+2];
        rdfNameString = args[i+3];
        i = i+4;
      }
      if (args.length >= i+4 && args[i].equalsIgnoreCase("-rdfdump")) {
        rdfDumpFileName = args[i+1];
        rdfDumpNamespaceString = args[i+2];
        rdfDumpNameString = args[i+3];
        i = i+4;
      }
      if (args.length >= i+1) {
        inputFileName = args[i];
        if (inputFileName.equals("-")) inputFileName = null;
        i++;
      }
      if (args.length >= i+1) {
        System.err.println("arguments ignored: " + args[i] + " ...");
      }
    }

    try {
      compileTripleProgram(inputFileName, OUT, "", "", false,
        rdfFileName, rdfNamespaceString, rdfNameString,
        rdfDumpFileName, rdfDumpNamespaceString, rdfDumpNameString);
    } catch (Exception e) {
      System.err.println("Error: " + e.getMessage());
      System.exit(1);
    }

  }
 
}

