/*
 * TripleTransform.java
 *
 * Created on January 29, 2002, 10:21 AM
 */

package org.semanticweb.triple.parser;

import java.util.*;
import org.semanticweb.triple.language.*;


/**
 * Translates LTerm program into a Triple program,
 * on a clause-by-clause basis to allow streaming
 * @author  sintek
 * @version 
 */
public class TripleTransform {
  
  public static Clause transformClause(LTerm lclause, LTerm lmodel) throws Exception {
    // note that this method is probably not entirely correct if
    // models are nested since the variables of outer models are not
    // respected!
    ArrayList modelVariables = new ArrayList();
    findFreeVariables(lmodel, modelVariables);
    if (lclause instanceof LStruct) {
      // lclause = [FORALL(vars)](head <- body)
      String functor = getFunctor2Symbol((LStruct)lclause);
      LTerm headAndBody = null;
      ArrayList variables = null; // of LVariable
      if (isFORALL(functor)) {
        variables = getFunctorArgs((LStruct)lclause);
        ArrayList args = ((LStruct)lclause).getArgs();
        if (args.size() == 1)
          headAndBody = (LTerm)args.get(0);
        else
          throw new ParseException("illegal clause: " + lclause);
      } else { // no FORALL
        headAndBody = lclause;
      }
      if (variables == null) variables = new ArrayList();
      findFreeVariables(headAndBody, variables, modelVariables);
      // System.out.println("FORALL vars: " + variables);
      return transformClause(variables, headAndBody, modelVariables, lmodel);
    } else { // clause is not a structure -> simple fact
      return transformClause(null, lclause, modelVariables, lmodel);
    }
  }

  public static Clause transformClause(ArrayList variables, LTerm headAndBody, 
      ArrayList mvariables, LTerm lmodel) throws Exception {
    Clause clause = new Clause();
    if (variables != null) {
      for (Iterator varIter = variables.iterator(); varIter.hasNext();) {
        LVariable var = (LVariable)varIter.next();
        clause.putVariables(new Variable(var.getVariable()));
      }
    }
    if (headAndBody instanceof LStruct) {
      if (isImpliedBy((LStruct)headAndBody)) { // rule or query: [h] <- b
        ArrayList args = ((LStruct)headAndBody).getArgs();
        if (args.size() == 1) // <- query
          clause.putBody(transformFormula((LTerm)args.get(0)));
        else { // rule: h <- b
          clause.putHead(transformFormula((LTerm)args.get(0)));
          clause.putBody(transformFormula((LTerm)args.get(1)));
        }
      } else { // fact or special directive
        Formula factTerm = transformFormula(headAndBody);
        if (isEXTERNAL(getFunctorSymbol((LStruct)headAndBody))) { // directive -> query
          // ... isDirective will also include IMPORT etc. ...
          clause.putBody(factTerm);
        } else {
          clause.putHead(factTerm);
        }
      }
    } else { // simple fact
      clause.putHead(transformFormula(headAndBody));
    }
    // model and model variables
    if (mvariables != null) {
      for (Iterator mvarIter = mvariables.iterator(); mvarIter.hasNext();) {
        LVariable mvar = (LVariable)mvarIter.next();
        clause.putModelVariables(new Variable(mvar.getVariable()));
      }
    }
    clause.putModel(transformTerm(lmodel));
    return clause;
  }
  
  public static Formula transformFormula(LTerm lformula) throws Exception {
    // check for: and, or, <-, ->, <->,
    // not, neg
    // triple/molecule, forall/exists,
    // @, and finally: Horn atom
    if (lformula instanceof LStruct) {
      String functor = getFunctorSymbol((LStruct)lformula);
      String functor2 = getFunctor2Symbol((LStruct)lformula);
      if (   isAND(functor)     || isOR(functor)
          || isIMPLIES(functor) || isIMPLIEDBY(functor) 
          || isEQUIV(functor)) {
        BinaryFormula binaryFormula = null;
        if      (isAND(functor)) binaryFormula = new And();
        else if (isOR(functor)) binaryFormula = new Or();
        else if (isIMPLIES(functor)) binaryFormula = new Implies();
        else if (isIMPLIEDBY(functor)) binaryFormula = new ImpliedBy();
        else if (isEQUIV(functor)) binaryFormula = new Equiv();
        ArrayList args = ((LStruct)lformula).getArgs();
        if (args.size() != 2)
          throw new ParseException("illegal formula: " + lformula);
        binaryFormula.putFirstFormula(transformFormula((LTerm)args.get(0)));
        binaryFormula.putSecondFormula(transformFormula((LTerm)args.get(1)));
        return binaryFormula;
      } else if (isNOT(functor) || isNEG(functor)) {
        UnaryFormula unaryFormula = null;
        if      (isNOT(functor)) unaryFormula = new Not();
        else if (isNEG(functor)) unaryFormula = new Neg();
        ArrayList args = ((LStruct)lformula).getArgs();
        if (args.size() != 1)
          throw new ParseException("illegal formula: " + lformula);
        unaryFormula.putFormula(transformFormula((LTerm)args.get(0)));
        return unaryFormula;
      } else if (isSB(functor2)) { // triple/molecule
        ArrayList f2args = ((LStruct)((LStruct)lformula).getFunctor()).getArgs();
        if (f2args.size() != 1)
          throw new ParseException("illegal formula: " + lformula);
        LTerm lsubject = (LTerm)f2args.get(0);
        ArrayList args = ((LStruct)lformula).getArgs();
        // only molecules are created, not triples ... !!!
        Molecule molecule = new Molecule();
        molecule.putSubject(transformTerm(lsubject));
        for (Iterator argIter = args.iterator(); argIter.hasNext();) {
          LTerm arg = (LTerm)argIter.next(); // must be p -> o
          if (arg instanceof LStruct) {
            String functorString = getFunctorSymbol((LStruct)arg);
            if (isIMPLIES(functorString)) {
              ArrayList predicateObject = ((LStruct)arg).getArgs();
              if (predicateObject.size() == 2) {
                LTerm predicate = (LTerm)predicateObject.get(0);
                LTerm object = (LTerm)predicateObject.get(1);
                molecule.putPredicateAndObject(
                  transformTerm(predicate),
                  transformTerm(object));
              } else
                throw new ParseException("illegal molecule: " + lformula);
            } else
              throw new ParseException("illegal molecule: " + lformula);
          } else
            throw new ParseException("illegal molecule: " + lformula);
        }
        return molecule;
      } else if (isFORALL(functor2) || isEXISTS(functor2)) {
        ArrayList variables = getFunctorArgs((LStruct)lformula);
        ArrayList args = ((LStruct)lformula).getArgs();
        LTerm innerFormula = null;
        if (args.size() != 1)
          throw new ParseException("illegal quantifier formula: " + lformula);
        innerFormula = (LTerm)args.get(0);
        QuantifierFormula qformula;
        if (isFORALL(functor2)) qformula = new ForallFormula();
        else                    qformula = new ExistsFormula();
        if (variables != null) {
          for (Iterator varIter = variables.iterator(); varIter.hasNext();) {
            LVariable var = (LVariable)varIter.next();
            qformula.putVariables(new Variable(var.getVariable()));
          }
        }
        qformula.putFormula(transformFormula(innerFormula));
        return qformula;
      } else if (isAT(functor)) { // @(f,m)
        ArrayList args = ((LStruct)lformula).getArgs();
        if (args.size() != 2)
          throw new ParseException("illegal @ formula: " + lformula);
        LTerm innerFormula = (LTerm)args.get(0);
        LTerm model = (LTerm)args.get(1);
        Formula formula = transformFormula(innerFormula);
        formula.putModel(transformTerm(model));
        return formula;
      } else { // simple Horn atom
        return transformHornAtom(lformula);
      }
    } else // no structure -> Horn atom (or error)
      return transformHornAtom(lformula);
  }
  
  public static Formula transformHornAtom(LTerm latom) throws Exception {
    HornAtom atom = new HornAtom();
    if (latom instanceof LStruct) {
      LTerm predicate = ((LStruct)latom).getFunctor();
      if (predicate instanceof LSymbol)
        atom.putPredicateSymbol(((LSymbol)predicate).getSymbol());
      else
        throw new ParseException("illegal formula: " + latom);
      ArrayList args = ((LStruct)latom).getArgs();
      if (args != null) {
        for (Iterator argIter = args.iterator(); argIter.hasNext();) {
          LTerm arg = (LTerm)argIter.next();
          atom.putArgs(transformTerm(arg));
        }
      }
    } else if (latom instanceof LSymbol) {
      String predicate = ((LSymbol)latom).getSymbol();
      atom.putPredicateSymbol(predicate);
    } else {
      throw new ParseException("illegal formula: " + latom);
    }
    return atom;
  }
  
  public static Term transformTerm(LTerm lterm) throws Exception {
    if (lterm instanceof LSymbol) { // -> Literal
      Literal literal = new Literal();
      literal.putName(((LSymbol)lterm).getSymbol());
      return literal;
    } else if (lterm instanceof LInteger) { // -> IntegerTerm
      return new IntegerTerm(((LInteger)lterm).getInteger());
    } else if (lterm instanceof LDouble) { // -> DoubleTerm
      return new DoubleTerm(((LDouble)lterm).getDouble());
    } else if (lterm instanceof LVariable) { // -> Variable
      Variable variable = new Variable();
      variable.putLabel(((LVariable)lterm).getVariable());
      return variable;
    } else if (lterm instanceof LStruct) {
      // -> triple, reified triple, resource, structure ...
      LTerm functor = ((LStruct)lterm).getFunctor();
      String functor1 = getFunctorSymbol((LStruct)lterm);
      String functor2 = getFunctor2Symbol((LStruct)lterm);
      if (isSB(functor2)) { // triple/molecule "term"
        ArrayList f2args = ((LStruct)((LStruct)lterm).getFunctor()).getArgs();
        if (f2args.size() != 1)
          throw new ParseException("illegal molecule: " + lterm);
        LTerm lsubject = (LTerm)f2args.get(0);
        ArrayList args = ((LStruct)lterm).getArgs();
        MoleculeTerm molecule = new MoleculeTerm();
        molecule.putSubject(transformTerm(lsubject));
        for (Iterator argIter = args.iterator(); argIter.hasNext();) {
          LTerm arg = (LTerm)argIter.next(); // must be p -> o
          if (arg instanceof LStruct) {
            String functorString = getFunctorSymbol((LStruct)arg);
            if (isIMPLIES(functorString)) {
              ArrayList predicateObject = ((LStruct)arg).getArgs();
              if (predicateObject.size() == 2) {
                LTerm predicate = (LTerm)predicateObject.get(0);
                LTerm object = (LTerm)predicateObject.get(1);
                molecule.putPredicateAndObject(
                  transformTerm(predicate),
                  transformTerm(object));
              } else
                throw new ParseException("illegal molecule: " + lterm);
            } else
              throw new ParseException("illegal molecule: " + lterm);
          } else
            throw new ParseException("illegal molecule: " + lterm);
        }
        return molecule;
      // else: reified triple, .... (@ expression for inner molecules?)
      } else if (isCOLON(functor1)) { // resource
        ArrayList args = ((LStruct)lterm).getArgs();
        if (args != null && args.size() == 2) {
          Resource resource = new Resource();
          resource.putNamespace(transformTerm((LTerm)args.get(0)));
          resource.putLocalName(transformTerm((LTerm)args.get(1)));
          return resource;
        } else
          throw new ParseException("illegal resource/term: " + lterm);
      } else if (functor instanceof LSymbol) {
        Structure structure = new Structure();
        structure.putFunctor(((LSymbol)functor).getSymbol());
        ArrayList args = ((LStruct)lterm).getArgs();
        if (args != null) {
          for (Iterator argIter = args.iterator(); argIter.hasNext();) {
            LTerm arg = (LTerm)argIter.next();
            structure.putArgs(transformTerm(arg));
          }
        }
        return structure;
      } else {
        throw new ParseException("illegal term: " + lterm);
      }
    } else {
      throw new ParseException("term nyi: " + lterm);
    }
  }
  
  
  // auxiliaries --------------------
  
  static void findFreeVariables(LTerm lterm, ArrayList variables) {
    // find all free variables in lterm
    findFreeVariables(lterm, variables, null);
  }
  
  static void findFreeVariables(LTerm lterm, ArrayList variables,
      ArrayList mvariables) {
    // add free variables to variables, where variables are the
    // toplevel clause variables and mvariables are all variables in
    // the default model for this clause 
    ArrayList qvariables = (ArrayList)variables.clone(); // quantified vars as copy
    if (mvariables != null)
      qvariables.addAll(mvariables);
    findFreeVariables2(lterm, qvariables, variables);
  }
  
  static void findFreeVariables2(LTerm lterm, ArrayList qvariables, 
      ArrayList fvariables) {
    // collect free variables in fvariables
    if (lterm instanceof LVariable) {
      if (!((LVariable)lterm).isAnonymous() &&
          !qvariables.contains(lterm) && !fvariables.contains(lterm))
        fvariables.add(lterm);
    } else if (lterm instanceof LStruct) { // recurse
      String functor2 = getFunctor2Symbol((LStruct)lterm);
      if (isFORALL(functor2) || isEXISTS(functor2)) {
        ArrayList quantVariables = getFunctorArgs((LStruct)lterm);
        qvariables = (ArrayList)qvariables.clone();
        qvariables.addAll(quantVariables);
        ArrayList args = ((LStruct)lterm).getArgs();
        for (Iterator argsIter = args.iterator(); argsIter.hasNext();) {
          LTerm arg = (LTerm)argsIter.next();
          findFreeVariables2(arg, qvariables, fvariables);
        }
      } else { // no exists/forall
        findFreeVariables2(((LStruct)lterm).getFunctor(), qvariables, fvariables);
        ArrayList args = ((LStruct)lterm).getArgs();
        for (Iterator argsIter = args.iterator(); argsIter.hasNext();) {
          LTerm arg = (LTerm)argsIter.next();
          findFreeVariables2(arg, qvariables, fvariables);
        }
      }
    }
  }
  
  static String getFunctorSymbol(LStruct structure) {
    // returns null if functor is not a symbol
    LTerm functor = structure.getFunctor();
    if (functor instanceof LSymbol)
      return (((LSymbol)functor).getSymbol());
    else
      return null;
  }

  static String getFunctor2Symbol(LStruct structure) {
    // f(...)(...) -> f
    // returns null if is not a symbol
    LTerm functor = structure.getFunctor();
    if (functor instanceof LStruct)
      return getFunctorSymbol((LStruct) functor);
    else
      return null;
  }
  
  static ArrayList getFunctorArgs(LStruct structure) {
    // f(args)(...) -> args
    LTerm functor = structure.getFunctor();
    return ((LStruct)functor).getArgs();
  }
  
  /*
  static String getUltimateFunctorSymbol(LStruct structure) {
    LTerm functor = structure.getFunctor();
    if (functor instanceof LStruct)
      return getUltimateFunctorSymbol((LStruct)functor);
    else if (functor instanceof LSymbol)
      return (((LSymbol)functor).getSymbol());
    else
      return null;
  }*/

  static boolean isMolecule(LStruct struct) {
    return getSBFunctor(struct) != null;
  }

  static LTerm getSBFunctor(LStruct struct) { // [](f)(...) -> f
    // returns null if struct is not a square brackets structure
    LTerm functor = struct.getFunctor();
    if (functor instanceof LStruct) {
      LTerm ffunctor = ((LStruct)functor).getFunctor();
      if (ffunctor instanceof LSymbol
          && isSB(((LSymbol)ffunctor).getSymbol())) {
        ArrayList args = ((LStruct)functor).getArgs();
        if (args.size() == 1)
          return (LTerm)args.iterator().next();
        else
          return null;
      } else
        return null;
    } else
      return null;
  }

  static LTerm getSingleArg(LStruct structure) throws ParseException {
    Collection args = structure.getArgs();
    if (args != null && args.size() == 1)
      return (LTerm)args.iterator().next();
    else
      throw new ParseException("one argument expected: " + structure);
  }

  static LTerm getFirstArg(LStruct structure) throws ParseException {
    Collection args = structure.getArgs();
    if (args != null && args.size() > 0)
      return (LTerm)args.iterator().next();
    else
      throw new ParseException("argument expected: " + structure);
  }

  static LTerm getSecondArg(LStruct structure) throws ParseException {
    Collection args = structure.getArgs();
    if (args != null && args.size() > 1) {
      Iterator argsIterator = args.iterator();
      argsIterator.next();
      return (LTerm)argsIterator.next();
    } else
      throw new ParseException("arguments expected: " + structure);
  }

  static boolean isResource(LStruct struct) { // :(n,l)
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isCOLON(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 2;
    }
    return false;
  }

  static boolean isImpliedBy(LStruct struct) { // [head] <- body
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isIMPLIEDBY(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 1 || args.size() == 2;
    }
    return false;
  }

  static boolean isAtFormula(LStruct struct) { // @(f,m)
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isAT(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 2;
    }
    return false;
  }

  static boolean isAndFormula(LStruct struct) { // f1 and f2
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isAND(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 2;
    }
    return false;
  }

  /*
  static boolean isClause(LStruct struct) { // CLAUSE(clause)
    LTerm functor = struct.getFunctor();
    if (functor instanceof LSymbol
        && isCLAUSE(((LSymbol)functor).getSymbol())) {
      ArrayList args = struct.getArgs();
      return args.size() == 1;
    }
    return false;
  }
  */


  // token tests

  static boolean isFORALL(String functor) {
    return "FORALL".equals(functor)
      || "forall".equals(functor);
  }
  
  static boolean isEXISTS(String functor) {
    return "EXISTS".equals(functor)
      || "exists".equals(functor);
  }
  
  static boolean isNOT(String functor) {
    return "NOT".equals(functor)
      || "not".equals(functor);
  }

  static boolean isNEG(String functor) {
    return "NEG".equals(functor)
      || "neg".equals(functor);
  }
  
  static boolean isEXTERNAL(String functor) {
    return "EXTERNAL".equals(functor)
      || "external".equals(functor);
  }
  
  static boolean isSB(String functor) {
    return "[]".equals(functor);
  }

  static boolean isCOLON(String functor) {
    return ":".equals(functor);
  }

  static boolean isAT(String functor) {
    return "@".equals(functor);
  }
  
  static boolean isIMPLIES(String functor) {
    return "->".equals(functor);
  }

  // we should distinguish between the different uses!!!
  static boolean isIMPLIEDBY(String functor) {
    return "<-".equals(functor)
      || ":-".equals(functor)
      || "?-".equals(functor);
  }
  
  static boolean isEQUIV(String functor) {
    return "<->".equals(functor);
  }

  static boolean isAND(String functor) {
    return "AND".equals(functor)
      || "and".equals(functor)
      || ",".equals(functor);
  }
  
  static boolean isOR(String functor) {
    return "OR".equals(functor)
      || "or".equals(functor)
      || ";".equals(functor);
  }
  
}
